# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Callable, Dict, Optional

import torch
from transformers import PretrainedConfig, PreTrainedModel


class BaseFoundationConfigMixin(PretrainedConfig, ABC):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class BaseFoundationModelMixin(PreTrainedModel):
    @abstractmethod
    def get_generation_position_id(self, **kwargs) -> Dict[str, torch.Tensor]:
        """ """
        pass

    def get_position_id_func(self) -> Optional[Callable]:
        return None
